#!/bin/sh
# vim600: set foldmethod=marker: # Faster folding in VIM.
#
# Shell script to test pwman. {{{1
# This is intended to be a generic shell testing framework.
#
# Copyright (C) 2007 Lorance Stinson All rights reserved.  This script is free
# software.  It may be copied or modified according to the same terms as Perl.
#
# For more information email <LoranceStinson+pwman@gmail.com>.
# Or see http://lorance.freeshell.org/pwman/

# Variables. {{{1

# Make the code a bit cleaner.
TAB="	"
YES="Yes"
NO="No"
ME="`basename $0`"

# Defaults.
program="pwman"         # The name of the program being tested.
quiet="$NO"             # Quiet mode. (Only failed tests.)
verbose="$NO"           # Verbose output.
onlylist="$NO"          # Only list the sections and names of the tests.
keepfiles="$NO"         # Keep the test files.

# These variables are used internally by run_test().
separator="$NO"         # Separator between tests in verbose mode.
fail_count="0"          # The number of test that failed.
pass_count="0"          # The number of test that passed.
skip_count="0"          # The number of tests skipped.
prev_section=""         # The previous section.
testout=""              # The output from the command.
testexit=0              # The exit code the command returned.
status=""               # The status of the text. Pass or Fail.
errexit=""              # Indicates if the exit code check failed.
errout=""               # Indicates if the output check failed.
sec_count="0"           # The number of sections encountered.
test_count="0"          # The number of tests in the current section.
sec_skip=""             # A list of sections to skip.
skip_sec="$NO"          # Skip the current section.

# These define each test.
# If any variable is set to an empty string it is ignored.
command="./$program"    # The command to run.
name=""                 # The name of the test. (Required)
description=""          # An option description for the test.
parameters=""           # Command line arguments passed to the program.
exitcode=""             # The expected exit code.
input=""                # Input to pipe to the program.
output=""               # The expected output of the test.

# Functions. {{{1

# usage() - Display the command line help text. {{{2
usage () {
    if [ "$1" ] ; then
        exec 1>&2
        echo "$1"
        echo ""
    fi
    cat<<EOHELP
Usage:  $ME [OPTIONS] [PROGRAM]
A shell utility for testing $program.
Options:
    -h      This help text.
    -k      Keep the test files. (test.dat, test.dat~, test.tsv)
            If the test files are present the tests will not run.
    -l      Only list the sections and names of the tests.
    -o FILE Write all output to FILE instead of STDOUT.
    -q      Quiet output. Print only the names of the tests that fail.
    -s      A comma separated list of section numbers to skip.
    -v      Verbose output. Overrides quiet.
EOHELP
    if [ "$1" ] ; then
        exit 1
    else
        exit 0
    fi
}

# run_test() - Run a test. {{{2
#
# These variables describe the test to be run.
# $name         The name of the test. (Required)
# $description  An option description for the test.
#               The description is only printed if the test fails or if
#               verbose mode is used.
# $command      The command that is run. (Required)
# $input        This is piped to the STDIN of the command.
# $parameters   Parameters passed to the command.
# $output       The output of the command is checked against this.  If the
#               output does not match this variable exactly the test fails.
#               STDIN and STDERR are captured together.
# $exitcode     The exit code of the command is checked against this. If
#               the exit code does not match exactly the test fails.
#
# These variables are used internally.
# $separator    Separator between tests in verbose mode. $YES or $NO.
# $fail_count   The number of test that failed.
# $pass_count   The number of test that passed.
# $skip_count   The number of tests skipped.
# $prev_section The previous section.
# $testout      The output from the command.
# $testexit     The exit code the command returned.
# $status       The status of the text. Pass or Fail.
# $errexit      Indicates if the exit code check failed. $YES or $NO.
# $errout       Indicates if the output check failed. $YES or $NO.
# $sec_count    The number of sections encountered.
# $test_count   The number of tests in the current section.
# $sec_skip     A list of sections to skip.
# $skip_sec     Skip the current section.
#
# For simplicity this script is not meant to handle lots of output. Temporary
# files and cmp or diff would work better in that case. Variables are used for
# flexibility and clarity.
run_test () {
    # Deal with the section if there is one and we are not being quiet.
    if [ "$quiet" = "$NO" -a "$section" != "$prev_section" ] ; then
        # Print a separator if this is not the first section..
        if [ ! -z "$prev_section" ] ; then
            echo ""
        fi

        # Increment the section count and reset the test count.
        sec_count="`expr $sec_count + 1`"
        test_count="0"

        # See if this section should be skipped.
        if [ "`echo "$sec_skip" | sed -n -e "/^$sec_count\$/p"`" != "" ] ; then
            echo "Skipping Section $sec_count: $section"
            skip_sec="$YES"
        else
            echo "Section $sec_count: $section"
            skip_sec="$NO"
            [ "$separator" = "$NO" ] && echo ""
        fi

        prev_section="$section"
    fi

    # Increment the test count.
    test_count="`expr $test_count + 1`"

    # See if the test should be skipped.
    if [ "$skip_sec" = "$YES" ] ; then
        echo "$sec_count.$test_count (Skip) $name"
        skip_count="`expr $skip_count + 1`"
        return 0
    fi

    # Print a separator line between tests in verbose mode.
    if [ "$verbose" = "$YES" ] ; then
        if [ "$separator" = "$YES" ] ; then
            echo ""
        fi
        separator="$YES"
    fi

    # Only print the name if that is what the user wants.
    if [ "$onlylist" = "$YES" ] ; then
        echo "$name"
        return
    fi

    # Default to pass.
    status="Pass"
    errexit="$NO"
    errout="$NO"

    # Run the test.
    testout="`echo "$input" | $command $parameters 2>&1`"
    testexit="$?"

    # Check the exit code.
    if [ "$exitcode" -a "$testexit" -ne "0$exitcode" ] ; then
        status="Fail"
        errexit="$YES"
    fi

    # Check the output.
    if [ "$output" -a "X$testout" != "X$output" ] ; then
        status="Fail"
        errout="$YES"
    fi

    # Print details if the test failed.
    if [ "$status" != "Pass" ] ; then
        echo "$sec_count.$test_count ($status) $name"
        if [ "$quiet" = "$NO" ] ; then
            # Optional description
            if [ "$description" ] ; then
                echo "Description: $description"
            fi

            # Print the command that was executed.
            echo "Command: $command $parameters"

            # Print the input if there was any.
            if [ "$input" ] ; then
                echo "Input: ---------------------------------------------------------------"
                echo "$input"
                echo "----------------------------------------------------------------------"
            fi

            # Print the exit code error if it failed.
            if [ "$errexit" = "$YES" ] ; then
                echo "Exit code: $testexit Expected: $exitcode"
            fi

            # Print the output error if it failed.
            if [ "$errout" = "$YES" ] ; then
                echo "Output was: ----------------------------------------------------------"
                echo "$testout"
                echo "Expected: ------------------------------------------------------------"
                echo "$output"
                echo "----------------------------------------------------------------------"
            fi
        fi

        # Add this to the failed count.
        fail_count="`expr $fail_count + 1`"
    elif [ "$quiet" = "$NO" ] ; then
        # Print the name and status.
        echo "$sec_count.$test_count ($status) $name"

        # Print the details if being verbose.
        if [ "$verbose" = "$YES" ] ; then
            # Optional description
            if [ "$description" ] ; then
                echo "Description: $description"
            fi

            # Print the command that was executed.
            echo "Command: $command $parameters"

            # Print the input if there was any.
            if [ "$input" ] ; then
                echo "Input: ---------------------------------------------------------------"
                echo "$input"
                echo "----------------------------------------------------------------------"
            fi

            # Print the exist code it was checked.
            if [ "$exitcode" ] ; then
                echo "Exit Code: $testexit"
            fi

            # Print the output if it was checked.
            if [ "$output" ] ; then
                echo "Output: --------------------------------------------------------------"
                echo "$testout"
                echo "----------------------------------------------------------------------"
            fi
        fi

        # Add this to the pass count.
        pass_count="`expr $pass_count + 1`"
    fi
}

# final_status() - Final status message. {{{2
# Does not print when in quiet mode.
final_status(){
    if [ "$quiet" = "$NO" -a "$onlylist" = "$NO" ] ; then
        echo ""
        (echo "Total$TAB`expr $pass_count + $fail_count + $skip_count`"
         echo "Passed$TAB$pass_count"
         echo "Failed$TAB$fail_count"
         if [ "$skip_count" -ne "0" ] ; then
             echo "Skipped$TAB$skip_count"
         fi
         ) |
         awk -F "$TAB" '{printf "%-6s : %4d\n", $1, $2}'
        if [ "$fail_count" -ne "0" ] ; then
            echo ""
            echo "*** Tests have failed! ***"
            echo "Please email the output of this program (without the -q option) to:"
            echo "LoranceStinson+awk_data@gmail.com"
        fi
    fi
}

# Check and handle commandline options. {{{1
while getopts ":hklo:qs:v" OPTION ; do
    case $OPTION in
        h)  usage;;                 # Print the help text.
        k)  keepfiles=$YES;;        # Only list the sections and names of the tests.
        l)  onlylist=$YES;;         # Only list the sections and names of the tests.
        o)  exec 1>$OPTARG;;        # Write all output to a file.
        q)  quiet=$YES;;            # Quite mode.
        s)  sec_skip="$OPTARG";;    # A comma separated list of section numbers to skip
        v)  verbose=$YES;;          # Verbose output.
        *)  if [ "$OPTARG" = "-" ] ; then
                usage # They probably only want help. Catches --help.
            else
                usage "Invalid option '$OPTARG'." # Illegal option.
            fi;;
    esac
done
shift `expr $OPTIND - 1`

# Make sure the test files do not exist.
if [ -f "test.dat" -o -f "test.tsv" ] ; then
    echo "Existing test files found." 1>&2
    echo "Aborting test." 1>&2
    exit 1
fi

# Verbose overrides quiet.
[ "$verbose" = "$YES" ] && quiet="$NO"

# Fix the sections to skip if any were given.
[ "$sec_skip" ] && sec_skip="`echo "$sec_skip" | tr ',' '\n'`"

# Is the command specified?
if [ "$#" -ne "0" ] ; then
    command=$1
    if [ ! -x "$command" ] ; then
        echo "'$command' is not executable." 1>&2
        exit 1
    fi
fi

# Tests. {{{1

# Initialization. {{{2
section="Initialization."

# Create the test database. {{{3
name="Create the test database."
parameters="-C -F test.dat"
exitcode=0
input="pass
pass"
output="             -----+++++*****##### WARNING #####*****+++++-----
This is a shell script. Shell scripts are not as secure as compiled programs.
If other programs can read the enviromnent variables of this one your master
password can be found with ease. It is possible your master password and other
passwords could be leaked. Use caution on multi user systems. If the shell used
to run this program does not have the commands test and echo as builtins your
password WILL BE AVAILABLE to ANYONE who happens to read the process table at
the right moment.

DO NOT PROCEED if you doubt the security of your passwords!

Master Password: 
Reenter the master Password: 
The database 'test.dat' has been created."
run_test

# Add accounts. {{{2
section="Add accounts."

# Add an account. (Account requested.) {{{3
name="Add an account. (Account requested.)"
parameters="-F test.dat -a"
exitcode=0
input="pass
account
name
password
password
Some notes."
output="Master Password: 
Account: User Name: Password: 
Reenter the Password: 
Notes: Account 'account' added."
run_test

# Add an account. (Account on command line.) {{{3
name="Add an account. (Account on command line.)"
parameters="-F test.dat -a foo"
exitcode=0
input="pass
bar
baz
baz
More notes."
output="Master Password: 
User Name: Password: 
Reenter the Password: 
Notes: Account 'foo' added."
run_test

# Add a duplicate account. {{{3
name="Add a duplicate account."
parameters="-F test.dat -a foo"
exitcode=0
input="pass
bar
baz
baz
More notes."
output="Master Password: 
User Name: Password: 
Reenter the Password: 
Notes: ERROR: That Account and User Name combination already exists.
ERROR: Account not added."
run_test

# List accounts. {{{2
section="List accounts."

# List all accounts. {{{3
name="List all accounts."
parameters="-F test.dat -l"
exitcode=0
input="pass"
output="Master Password: 
Account - User Name > Notes
account - name > Some notes.
foo - bar > More notes."
run_test

# List all accounts with passwords. {{{3
name="List all accounts with passwords."
parameters="-F test.dat -p -l"
exitcode=0
input="pass"
output="Master Password: 
Account - User Name (Password) > Notes
account - name (password) > Some notes.
foo - bar (baz) > More notes."
run_test

# List a single account with password. {{{3
name="List a single account with password."
parameters="-F test.dat -p -l foo"
exitcode=0
input="pass"
output="Master Password: 
Account - User Name (Password) > Notes
foo - bar (baz) > More notes."
run_test

# List a single account ignoring case. {{{3
name="List a single account ignoring case."
parameters="-F test.dat -i -l FOO"
exitcode=0
input="pass"
output="Master Password: 
Account - User Name > Notes
foo - bar > More notes."
run_test

# Search accounts. {{{2
section="Search accounts."

# Search for an account. {{{3
name="Search for an account."
parameters="-F test.dat -s bar"
exitcode=0
input="pass"
output="Master Password: 
Account - User Name > Notes
foo - bar > More notes."
run_test

# Search for an account with password. {{{3
name="Search for an account with password."
parameters="-F test.dat -p -s bar"
exitcode=0
input="pass"
output="Master Password: 
Account - User Name (Password) > Notes
foo - bar (baz) > More notes."
run_test

# Search for an account ignoring case. {{{3
name="Search for an account ignoring case."
parameters="-F test.dat -i -s BAR"
exitcode=0
input="pass"
output="Master Password: 
Account - User Name > Notes
foo - bar > More notes."
run_test

# Manipulate the database. {{{2
section="Manipulate the database."

# Change the master password. {{{3
name="Change the master password."
parameters="-F test.dat -c"
exitcode=0
input="pass
newpass
newpass"
output="Master Password: 
Enter the new Master Password: 
Reenter the new Master Password: 
The master password has been updated."
run_test

# Delete an account. {{{3
name="Delete an account."
parameters="-F test.dat -d foo"
exitcode=0
input="newpass"
output="Master Password: 
Account 'foo' deleted."
run_test

# Edit an account. {{{3
name="Edit an account."
parameters="-F test.dat -e account"
exitcode=0
input="newpass
ACCOUNT
NAME
PASSWORD
PASSWORD
New Notes."
output="Master Password: 
Account [account]: User Name [name]: Password [Enter to keep]: 
Reenter the new Password: 
> Some notes.
Notes [Enter to keep]: Account 'ACCOUNT' changed."
run_test

# Import/Export. {{{2
section="Import/Export."

# Import from a file. {{{3
echo "ACCOUNT	USER	PASS	NOTES." > test.tsv
name="Import from a file."
parameters="-F test.dat -I test.tsv"
exitcode=0
input="newpass"
output="Master Password: 
The file has been imported."
run_test

# Import from STDIN. {{{3
name="Import from STDIN."
parameters="-F test.dat -I -"
exitcode=0
input="newpass
account	user_name	password	Notes!"
output="Master Password: 
The file has been imported."
run_test

# Export to STDOUT. {{{3
name="Export to STDOUT."
parameters="-F test.dat -E -"
exitcode=0
input="newpass"
output="Master Password: 
ACCOUNT	NAME	PASSWORD	New Notes.
ACCOUNT	USER	PASS	NOTES.
account	user_name	password	Notes!"
run_test

# Export to a file. {{{3
# This should be the last test since $command is altered.
rm -f test.tsv > /dev/null 2>&1
echo "newpass" | $command -F test.dat -E test.tsv > /dev/null 2>&1
name="Export to a file."
command="cat"
parameters="test.tsv"
exitcode=0
input=""
output="ACCOUNT	NAME	PASSWORD	New Notes.
ACCOUNT	USER	PASS	NOTES.
account	user_name	password	Notes!"
run_test

# Remove the test files. {{{1
if [ "$keepfiles" = "$NO" ] ; then
    if [ "$verbose" = "$YES" ] ; then
        echo ""
        echo "Removing test files."
    fi
    rm -f test.dat* test.tsv > /dev/null 2>&1
fi

# Print a final status message. {{{1
final_status
